#include "in2mcmt.h"
#include "node.h"
#include "parser.h"

#include <stdlib.h>

#define KIND_INDEX_SET  1
#define KIND_SIZE       2
#define KIND_PARAM      3
#define KIND_PARAM_SET  4
#define KIND_TOKEN      5
#define KIND_TOKEN_SET  6

#define TYPE_BOOL  1
#define TYPE_INT   2

#define CONN_CASE_1   10
#define CONN_CASE_2   20
#define CONN_CASE_3   30
#define CONN_CASE_4a  41
#define CONN_CASE_4b  42
#define CONN_CASE_5   50
#define CONN_CASE_6   60
#define CONN_CASE_7   70
#define CONN_CASE_8   80
#define CONN_CASE_9   90

#define INDEX_VAR_X 1
#define INDEX_VAR_Y 2
#define INDEX_CONST 3
#define INDEX_OTHER 4

#define MAX_ID_LEN 1024

#define CASE_OTHER "___other"

/* struct to hold info about an MCMT variable */
struct mcmt_var {
    char mcmt_name[MAX_ID_LEN + 10];
    char orig_name[MAX_ID_LEN + 1];
    int kind;
    int type;
    int index_set;
    int size_var;
    int is_src;
    int is_dst;
};

int conn_count = 0;
int reconf_count = 0;
int mcmt_trans_count = 0;

/* struct to hold info about a case index in a reconfiguration transition */
struct case_index
{
    int type;
    char orig_name[MAX_ID_LEN + 1];
    int constant;
};

/* struct to hold info about a case in a reconfiguration transition */
struct reconf_case
{
    struct case_index *index;
    GArray *updates; // elements are of type GString *
};

void rewrite_boolean_literals(node_ptr *node);
void get_mcmt_vars(node_ptr node, GHashTable *src, GHashTable *dst,
        GArray *vars);
void get_index_sets(node_ptr node, GArray *vars);
void get_other_vars(node_ptr node, GHashTable *src, GHashTable *dst,
        GArray *vars);
void find_index_sets(node_ptr node, GArray *index_sets);
void find_ids_used_in_size(node_ptr node, GHashTable *ids);
void find_other_vars(node_ptr node, GArray *var_nodes);
int get_index_of_var(GArray *vars, char *orig_name);
int get_index_of_var_mcmt_name(GArray *vars, char *mcmt_name);
void output_src_dst_comment(GHashTable *src, GHashTable *dst);
void output_index_spec();
void output_variables(GArray *vars);

void printf_redir(GString *out, const char *format, ...);

void output_initial_formula(node_ptr node, GArray *vars);
node_ptr find_initial_formula(node_ptr node);
void translate_initial_formula(node_ptr node, GArray *vars, node_ptr fax,
        node_ptr fay);
void translate_initial_invariant_formulas(node_ptr node, GArray *vars);
void find_invariant_formulas(node_ptr node, GArray *invars, int in_invars);
void output_initial_token_constraints(GArray *vars);
void translate_initial_in(node_ptr node, GArray *vars, int val);
int is_index_set_spec(node_ptr node, GArray *vars);
void translate_initial_index_set_spec(node_ptr node, GArray *vars);
int is_index_set_rel(node_ptr node, GArray *vars);
void translate_initial_index_set_rel(node_ptr node, GArray *vars);
int translate_initial_set_of_indexes(node_ptr node);
void find_ints_in_cons(node_ptr node, GHashTable *elems);
void translate_initial_term(node_ptr node, GArray *vars, node_ptr fax,
        node_ptr fay);
void find_different_indexes_in_term(node_ptr node, GArray *indexes);
int indexes_equal(node_ptr i1, node_ptr i2);
void translate_id(node_ptr node, GArray *vars, node_ptr repl_x, node_ptr repl_y,
        GString *out);
void translate_initial_index(node_ptr node, GArray *vars);
void translate_size(node_ptr node, GArray *vars, GString *out);
void translate_initial_term_replacing_indexes(node_ptr node, GArray *vars,
        node_ptr repl_x, node_ptr repl_y);

void output_unsafe_formula(GArray *vars);

void output_communication_transitions(node_ptr node, GArray *vars);
void find_connections(node_ptr node, GArray *conns);
void translate_connection(node_ptr node, node_ptr tree, GArray *vars);
int determine_connection_case(node_ptr node);
void translate_connection_case_1(node_ptr node, node_ptr tree, GArray *vars);
void translate_connection_case_2(node_ptr node, node_ptr tree, GArray *vars);
void translate_connection_case_3(node_ptr node, node_ptr tree, GArray *vars);
void translate_connection_case_4a(node_ptr node, node_ptr tree, GArray *vars);
void translate_connection_case_4b(node_ptr node, node_ptr tree, GArray *vars);
void translate_connection_case_5(node_ptr node, node_ptr tree, GArray *vars);
void translate_connection_case_6(node_ptr node, node_ptr tree, GArray *vars);
void translate_connection_case_7(node_ptr node, node_ptr tree, GArray *vars);
void translate_connection_case_8(node_ptr node, node_ptr tree, GArray *vars);
void translate_connection_case_9(node_ptr node, node_ptr tree, GArray *vars);
void find_conjuncts(node_ptr node, GArray *conjs);
void translate_atom_replacing_ids_and_nums(node_ptr node, GArray *vars,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y, GString *out);
void translate_in_replacing_ids_and_nums(node_ptr node, GArray *vars, int val,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y, GString *out);
void translate_term_replacing_ids_and_nums(node_ptr node, GArray *vars,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y, GString *out);
node_ptr find_subcomp_guard(node_ptr node, char *name);
void output_upds_set_one_to_true(GArray *vars, char *mcmt_name);

void output_reconfiguration_transitions(node_ptr node, GArray *vars);
void find_reconfigurations(node_ptr node, GArray *reconfs, int in_reconf);
void translate_reconfiguration(node_ptr node, node_ptr tree, GArray *vars);
void find_component_guards(node_ptr node, GArray *guards);
void find_simple_indices_sorted(node_ptr node, GSList *indices);
gint compare_case_indexes(gconstpointer ci1, gconstpointer ci2);
void find_reconf_different_indexes(node_ptr node, GArray *vars,
        GArray *indexes, int in_set);
void translate_simple_reconfiguration(node_ptr node, node_ptr tree,
        GArray *vars, GSList *case_indices);
node_ptr rewrite_formula_replacing_var(node_ptr node, char *x, node_ptr n);
int is_update_atom(node_ptr node);
void translate_reconf_add_remove(int is_add, node_ptr par, node_ptr val,
        GArray* vars, GSList *cases,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y);
void translate_reconf_par_update(node_ptr par, node_ptr val,
        GArray* vars, GSList *cases,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y);
void translate_reconf_unset_parameters(GArray* vars, GSList *cases,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y);
void translate_reconf_unset_tokens(node_ptr tree, GArray* vars, GSList *cases);
void translate_reconf_invars(node_ptr tree, GArray* vars, GSList *cases);
void translate_conjunction_using_next_values(int structure, node_ptr node,
        GArray* vars, GSList *cases, GString *out);
struct reconf_case *get_reconf_case_for_const_index_node(node_ptr idx,
        GSList *cases);
void translate_in_using_next_values(node_ptr node, GArray* vars, GSList *cases,
        int val, GString *out);
void translate_atom_using_next_values(node_ptr node, GArray* vars,
        GSList *cases, GString *out);
void translate_size_using_next_values(node_ptr node, GArray* vars,
        GSList *cases, GString *out);
void translate_id_using_next_values(node_ptr node, GArray* vars,
        GSList *cases, GString *out);
void translate_reconf_cases(GSList *cases);
void get_elements_of_set(node_ptr node, GArray *elems);
void set_update_val(char *name, struct case_index *ci, GString *value,
        GSList *cases, GArray *vars);
struct case_index *get_case_index_for_node(node_ptr n,
        node_ptr repl_x, node_ptr repl_y, GSList *cases);

void input_to_mcmt(node_ptr tree, GHashTable *src, GHashTable *dst)
{
    //TODO: check identifers don't start with _  (M0)
    //TODO: check identifers len <= MAX_ID_LEN
    //TODO: check M1
    //TODO: check M4a

    rewrite_boolean_literals(&tree);

    /* MCMT variables */
    GArray *vars = g_array_new(FALSE, FALSE, sizeof(struct mcmt_var));

    get_mcmt_vars(tree, src, dst, vars);

    output_src_dst_comment(src, dst);
    output_index_spec();
    output_variables(vars);
    output_initial_formula(tree, vars);
    output_unsafe_formula(vars);
    output_communication_transitions(tree, vars);
    output_reconfiguration_transitions(tree, vars);

    g_array_free(vars, TRUE);
}

void rewrite_boolean_literals(node_ptr *node)
{
    node_ptr l;
    if (*node == NULL)
        return;

    switch(node_get_type(*node)) {
        case INVARIANTS:
            //fall through
        case INIT:
            //fall through
        case SUBCOMPONENTS:
            //fall through
        case CONNECTIONS:
            //fall through
        case RECONFIGURATIONS:
            //fall through
        case CONS:
            //fall through
        case AND:
            //fall through
        case OR:
            //fall through
        case IFF:
            //fall through
        case IMPLIES:
            //fall through
        case GUARD:
            rewrite_boolean_literals(node_get_left_ptr(*node));
            rewrite_boolean_literals(node_get_right_ptr(*node));
            break;
        case EXISTS:
            //fall through
        case FORALL:
            //fall through
        case SUBCOMPONENT:
            rewrite_boolean_literals(node_get_right_ptr(*node));
            break;
        case CONNECTION:
            rewrite_boolean_literals(node_get_left_ptr(node_get_right(*node)));
            break;
        case NOT:
            l = node_get_left(*node);

            if (node_get_type(l) == ID || node_get_type(l) == INDEXED) {
                *node = new_node(EQUAL, l, new_node(BOOL_FALSE, NULL, NULL));
            } else if (node_get_type(l) == NOT) {
                fprintf(stderr, "double negation\n");
                exit(1);
            }
            break;
        case ID:
            // fall through
        case INDEXED:
                *node = new_node(EQUAL, *node, new_node(BOOL_TRUE, NULL, NULL));
            break;
        default:
            break;
    }
}

void get_mcmt_vars(node_ptr node, GHashTable *src, GHashTable *dst,
        GArray *vars)
{
    get_index_sets(node, vars);
    get_other_vars(node, src, dst, vars);
}

void get_index_sets(node_ptr node, GArray *vars)
{
    int i;
    node_ptr n;
    char *s;
    struct mcmt_var v;
    GArray *is_nodes = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GHashTable *in_size = g_hash_table_new(g_str_hash, g_str_equal);

    find_index_sets(node, is_nodes);
    find_ids_used_in_size(node, in_size);

    for (i = 0; i < is_nodes->len; i++) {
        n = g_array_index(is_nodes, node_ptr, i);

        s = (char*) node_get_left(node_get_left(n));
        strcpy(v.mcmt_name, "_");
        strcat(v.mcmt_name, s);
        strcpy(v.orig_name, s);
        v.kind = KIND_INDEX_SET;
        v.type = TYPE_BOOL;
        v.index_set = -1;
        if (g_hash_table_contains(in_size, v.orig_name))
            v.size_var = vars->len + 1;
        else
            v.size_var = -1;
        v.is_src = 0;
        v.is_dst = 0;

        g_array_append_val(vars, v);

        if (v.size_var != -1) {
            strcpy(v.mcmt_name, "__size_");
            strcat(v.mcmt_name, v.orig_name);
            strcpy(v.orig_name, "");
            v.kind = KIND_SIZE;
            v.type = TYPE_INT;
            v.index_set = vars->len - 1;
            v.size_var = -1;
            v.is_src = 0;
            v.is_dst = 0;

            g_array_append_val(vars, v);
        }
    }

    g_array_free(is_nodes, TRUE);
    g_hash_table_destroy(in_size);
}

void find_index_sets(node_ptr node, GArray *index_sets)
{
    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case PARAMETER:
            if (node_get_type(node_get_right(node)) == INDEX_SET_TYPE)
                g_array_append_val(index_sets, node);
            break;
        case ID:
            // fall through
        case NUMBER:
            // fall through
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            break;
        default:
            find_index_sets(node_get_right(node), index_sets);
            find_index_sets(node_get_left(node), index_sets);
    }
}

void find_ids_used_in_size(node_ptr node, GHashTable *ids)
{
    char *id;

    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case SIZE:
            id = (char *) node_get_left(node_get_left(node));
            g_hash_table_add(ids, id);
            break;
        case ID:
            // fall through
        case NUMBER:
            // fall through
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            break;
        default:
            find_ids_used_in_size(node_get_left(node), ids);
            find_ids_used_in_size(node_get_right(node), ids);
            break;
    }
}

void get_other_vars(node_ptr node, GHashTable *src, GHashTable *dst,
        GArray *vars)
{
    int i;
    node_ptr n, nn, nnn;
    char *s;
    struct mcmt_var v;
    GArray *var_nodes = g_array_new(FALSE, FALSE, sizeof(node_ptr));

    find_other_vars(node, var_nodes);
    for (i = 0; i < var_nodes->len; i++) {
        n = g_array_index(var_nodes, node_ptr, i);

        if (node_get_type(n) == PARAMETER) {
            s = (char*) node_get_left(node_get_left(n));
            strcpy(v.mcmt_name, "_");
            strcat(v.mcmt_name, s);
            strcpy(v.orig_name, s);
            nn = node_get_right(n);
            if (node_get_type(nn) == SET_TYPE) {
                v.kind = KIND_PARAM_SET;
                nnn = node_get_left(nn);
                if (node_get_type(node_get_right(nnn)) == BOOLEAN)
                    v.type = TYPE_BOOL;
                else
                    v.type = TYPE_INT;

                s = (char*) node_get_left(node_get_left(nnn));
                v.index_set = get_index_of_var(vars, s);
                v.size_var = -1;
            } else {
                v.kind = KIND_PARAM;
                if (node_get_type(nn) == BOOLEAN)
                    v.type = TYPE_BOOL;
                else
                    v.type = TYPE_INT;
                v.index_set = -1;
                v.size_var = -1;
            }

            v.is_src = 0;
            v.is_dst = 0;
        } else {
            nn = node_get_left(n);
            s = (char*) node_get_left(node_get_left(nn));
            strcpy(v.mcmt_name, "_");
            strcat(v.mcmt_name, s);
            strcat(v.mcmt_name, "_token");
            strcpy(v.orig_name, s);
            nnn = node_get_right(nn);
            if (node_get_type(nnn) == SINGLE) {
                v.kind = KIND_TOKEN;
                v.index_set = -1;
            } else {
                v.kind = KIND_TOKEN_SET;
                s = (char*) node_get_left(nnn);
                v.index_set = get_index_of_var(vars, s);
            }
            v.type = TYPE_BOOL;
            v.size_var = -1;
            v.is_src = g_hash_table_contains(src, v.orig_name);
            v.is_dst = g_hash_table_contains(dst, v.orig_name);
        }

        g_array_append_val(vars, v);
    }

    g_array_free(var_nodes, TRUE);
}

void find_other_vars(node_ptr node, GArray *var_nodes)
{
    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case PARAMETER:
            if (node_get_type(node_get_right(node)) != INDEX_SET_TYPE)
                g_array_append_val(var_nodes, node);
            break;
        case SUBCOMPONENT:
                g_array_append_val(var_nodes, node);
            break;
        case ID:
            // fall through
        case NUMBER:
            // fall through
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            break;
        default:
            find_other_vars(node_get_right(node), var_nodes);
            find_other_vars(node_get_left(node), var_nodes);
    }
}

int get_index_of_var(GArray *vars, char *orig_name)
{
    int i;
    struct mcmt_var v;

    for (i = 0; i < vars->len; i++) {
        v = g_array_index(vars, struct mcmt_var, i);
        if (!strcmp(v.orig_name, orig_name))
            return i;
    }

    return -1;
}

int get_index_of_var_mcmt_name(GArray *vars, char *mcmt_name)
{
    int i;
    struct mcmt_var v;

    for (i = 0; i < vars->len; i++) {
        v = g_array_index(vars, struct mcmt_var, i);
        if (!strcmp(v.mcmt_name, mcmt_name))
            return i;
    }

    return -1;
}

void output_src_dst_comment(GHashTable *src, GHashTable *dst)
{
    GHashTableIter iter;
    gpointer key, value;

    printf(":comment Source components:");
    g_hash_table_iter_init(&iter, src);
    while (g_hash_table_iter_next(&iter, &key, &value))
        printf(" %s", (char *) key);
    printf("\n");

    printf(":comment Destination components:");
    g_hash_table_iter_init(&iter, dst);
    while (g_hash_table_iter_next(&iter, &key, &value))
        printf(" %s", (char *) key);
    printf("\n");
}

void output_index_spec()
{
    printf("\n");
    printf(":index int\n");
}

void output_variables(GArray *vars)
{
    int i;
    struct mcmt_var v;

    printf("\n");
    for (i = 0; i < vars->len; i++) {
        v = g_array_index(vars, struct mcmt_var, i);
        if (v.kind == KIND_SIZE || v.kind == KIND_PARAM || v.kind == KIND_TOKEN)
            printf(":global");
        else
            printf(":local");
        printf(" %s", v.mcmt_name);
        if (v.type == TYPE_BOOL)
            printf(" bool\n");
        else
            printf(" int\n");
    }
}

void printf_redir(GString *out, const char *fmt, ...)
{
    va_list args;
    va_start(args, fmt);
    if (out)
        g_string_append_vprintf(out, fmt, args);
    else
        vprintf(fmt, args);
    va_end(args);
}

node_ptr find_initial_formula(node_ptr node)
{
    node_ptr result;

    if (node == NULL)
        return NULL;

    switch(node_get_type(node)) {
        case INIT:
            result = node_get_left(node);
            break;
        case ID:
            // fall through
        case NUMBER:
            // fall through
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            result = NULL;
            break;
        default:
            result = find_initial_formula(node_get_left(node));
            if (!result)
                result = find_initial_formula(node_get_right(node));
    }

    return result;
}

void output_initial_formula(node_ptr node, GArray *vars)
{
    printf("\n");
    printf(":initial\n");
    printf(":var x\n");
    printf(":var y\n");
    printf(":cnj");
    translate_initial_formula(find_initial_formula(node), vars, Nil, Nil);
    translate_initial_invariant_formulas(node, vars);
    output_initial_token_constraints(vars);
    printf("\n");
}

void translate_initial_formula(node_ptr node, GArray *vars, node_ptr fax,
        node_ptr fay)
{
    node_ptr id, set_id;
    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case FORALL:
            if (fay != Nil) {
                fprintf(stderr, "Max 2 nested forall in initial formula\n");
                exit(1);
            }
            id = node_get_left(node_get_left(node));
            set_id = node_get_right(node_get_left(node));
            if (fax == Nil)
                fax = id;
            else
                fay = id;
            printf(" (=>");
            if (node_get_type(node_get_left(node)) == SCOPE_COMPLEMENT)
                printf(" (not");
            translate_initial_term(set_id, vars, fax, fay);
            if (fay == Nil)
                printf("[x]");
            else
                printf("[y]");
            if (node_get_type(node_get_left(node)) == SCOPE_COMPLEMENT)
                printf(")");
            translate_initial_formula(node_get_right(node), vars, fax, fay);
            printf(")");
            break;
        case AND:
            printf(" (and");
            translate_initial_formula(node_get_left(node), vars, fax, fay);
            translate_initial_formula(node_get_right(node), vars, fax, fay);
            printf(")");
            break;
        case OR:
            printf(" (or");
            translate_initial_formula(node_get_left(node), vars, fax, fay);
            translate_initial_formula(node_get_right(node), vars, fax, fay);
            printf(")");
            break;
        case IMPLIES:
            printf(" (=>");
            translate_initial_formula(node_get_left(node), vars, fax, fay);
            translate_initial_formula(node_get_right(node), vars, fax, fay);
            printf(")");
            break;
        case IFF:
            printf(" (=");
            translate_initial_formula(node_get_left(node), vars, fax, fay);
            translate_initial_formula(node_get_right(node), vars, fax, fay);
            printf(")");
            break;
        case NOT:
            translate_initial_in(node_get_left(node), vars, 0);
            break;
        case IN:
            translate_initial_in(node, vars, 1);
            break;
        case EQUAL:
            if (is_index_set_spec(node, vars))
                translate_initial_index_set_spec(node, vars);
            else if (is_index_set_rel(node, vars))
                translate_initial_index_set_rel(node, vars);
            else
                translate_initial_term(node, vars, fax, fay);
            break;
        case LE:
            if (is_index_set_rel(node, vars))
                translate_initial_index_set_rel(node, vars);
            else
                translate_initial_term(node, vars, fax, fay);
            break;
        case GE:
            // fall through
        case LT:
            // fall through
        case GT:
            // fall through
        case NOTEQUAL:
            translate_initial_term(node, vars, fax, fay);
            break;
        case ID:
            // fall through
        case NUMBER:
            break;
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            translate_initial_term(node, vars, fax, fay);
            break;
        default:
            break;
    }
}

void translate_initial_invariant_formulas(node_ptr node, GArray *vars)
{
    int i;
    node_ptr n;
    GArray *invars = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    find_invariant_formulas(node, invars, 0);
    for (i = 0; i < invars->len; i++) {
        n = g_array_index(invars, node_ptr, i);
        translate_initial_formula(n, vars, Nil, Nil);
    }

    g_array_free(invars, TRUE);
}

void find_invariant_formulas(node_ptr node, GArray *invars, int in_invars)
{
    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case INVARIANTS:
            find_invariant_formulas(node_get_left(node), invars, 1);
            break;
        case CONS:
            find_invariant_formulas(node_get_right(node), invars, in_invars);
            find_invariant_formulas(node_get_left(node), invars, in_invars);
            break;
        case ID:
            // fall through
        case NUMBER:
            break;
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            if (in_invars)
                g_array_append_val(invars, node);
            break;
        default:
            if (in_invars) {
                g_array_append_val(invars, node);
            } else {
                find_invariant_formulas(node_get_right(node), invars, 0);
                find_invariant_formulas(node_get_left(node), invars, 0);
            }
            break;
    }
}

void output_initial_token_constraints(GArray *vars)
{
    int i;
    struct mcmt_var v;

    for (i = 0; i < vars->len; i++) {
        v = g_array_index(vars, struct mcmt_var, i);
        if (v.kind == KIND_TOKEN)
            printf(" (= %s", v.mcmt_name);
        if (v.kind == KIND_TOKEN_SET)
            printf(" (= %s[x]", v.mcmt_name);
        if (v.kind == KIND_TOKEN || v.kind == KIND_TOKEN_SET) {
            if (v.is_src)
                printf(" true)");
            else
                printf(" false)");
        }
    }
}

void translate_initial_in(node_ptr node, GArray *vars, int val)
{
    node_ptr term;
    node_ptr set;
    int i;
    struct mcmt_var v;

    term = node_get_left(node);
    set = node_get_right(node);
    i = get_index_of_var(vars, (char *) node_get_left(set));
    v = g_array_index(vars, struct mcmt_var, i);

    printf(" (=> (= x");
    translate_initial_term(term, vars, Nil, Nil);
    printf(") (= %s[x]", v.mcmt_name);
    if (val)
        printf(" true))");
    else
        printf(" false))");
}

int is_index_set_spec(node_ptr node, GArray *vars)
{
    node_ptr set, val;
    int i;
    struct mcmt_var v;

    if (node_get_type(node) != EQUAL)
        return 0;

    set = node_get_left(node);
    val = node_get_right(node);
    if (set == NULL || node_get_type(set) != ID) {
        set = node_get_right(node);
        val = node_get_left(node);
    }

    if (set == NULL || node_get_type(set) != ID)
        return 0;

    i = get_index_of_var(vars, (char *) node_get_left(set));
    v = g_array_index(vars, struct mcmt_var, i);
    if (v.kind != KIND_INDEX_SET)
        return 0;
    if (val != NULL && node_get_type(val) != CONS)
        return 0;

    return 1;
}

void translate_initial_index_set_spec(node_ptr node, GArray *vars)
{
    node_ptr set, val;
    int i, size;
    struct mcmt_var v, v_size;

    set = node_get_left(node);
    val = node_get_right(node);
    if (set == NULL || node_get_type(set) != ID) {
        set = node_get_right(node);
        val = node_get_left(node);
    }

    i = get_index_of_var(vars, (char *) node_get_left(set));
    v = g_array_index(vars, struct mcmt_var, i);

    if (v.size_var != -1) {
        printf(" (and");
    }

    printf(" (= %s[x]", v.mcmt_name);
    size = translate_initial_set_of_indexes(val);
    printf(")");

    if (v.size_var != -1) {
        v_size = g_array_index(vars, struct mcmt_var, v.size_var);
        printf(" (= %s %d))", v_size.mcmt_name, size);
    }
}

int is_index_set_rel(node_ptr node, GArray *vars)
{
    node_ptr l, r;
    int i;
    struct mcmt_var v;

    if (node_get_type(node) != EQUAL && node_get_type(node) != LE)
        return 0;

    l = node_get_left(node);
    r = node_get_right(node);
    if (l == NULL || node_get_type(l) != ID
            || r == NULL || node_get_type(r) != ID)
        return 0;

    i = get_index_of_var(vars, (char *) node_get_left(l));
    v = g_array_index(vars, struct mcmt_var, i);
    if (v.kind != KIND_INDEX_SET)
        return 0;

    i = get_index_of_var(vars, (char *) node_get_left(r));
    v = g_array_index(vars, struct mcmt_var, i);
    if (v.kind != KIND_INDEX_SET)
        return 0;

    return 1;
}

void translate_initial_index_set_rel(node_ptr node, GArray *vars)
{
    node_ptr l, r;
    int i;
    struct mcmt_var v_l, v_size_l, v_r, v_size_r;

    l = node_get_left(node);
    r = node_get_right(node);

    i = get_index_of_var(vars, (char *) node_get_left(l));
    v_l = g_array_index(vars, struct mcmt_var, i);

    i = get_index_of_var(vars, (char *) node_get_left(r));
    v_r = g_array_index(vars, struct mcmt_var, i);

    if (v_l.size_var != -1 && v_r.size_var != -1)
        printf(" (and");

    if (node_get_type(node) == EQUAL)
        printf(" (= ");
    else
        printf(" (=> ");
    printf("%s[x] %s[x])", v_l.mcmt_name, v_r.mcmt_name);

    if (v_l.size_var != -1 && v_r.size_var != -1) {
        v_size_l = g_array_index(vars, struct mcmt_var, v_l.size_var);
        v_size_r = g_array_index(vars, struct mcmt_var, v_r.size_var);
        if (node_get_type(node) == EQUAL)
            printf(" (= ");
        else
            printf(" (<= ");
        printf("%s %s))", v_size_l.mcmt_name, v_size_r.mcmt_name);
    }
}

int translate_initial_set_of_indexes(node_ptr node)
{
    GHashTable *elems = g_hash_table_new(g_int_hash, g_int_equal);
    GHashTableIter iter;
    gpointer key, val;
    int i, size;
    
    if (node == NULL) {
        printf(" false");
        return 0;
    }

    find_ints_in_cons(node, elems);
    size = g_hash_table_size(elems);
    for (i = 0; i < size - 1; i++)
        printf(" (or");
    g_hash_table_iter_init(&iter, elems);
    i = 0;
    while(g_hash_table_iter_next(&iter, &key, &val)) {
        printf(" (= x %d)", *((int *)key));
        free(key);
        if (i > 0)
            printf(")");
        i++;
    }

    g_hash_table_destroy(elems);

    return size;
}

void find_ints_in_cons(node_ptr node, GHashTable *elems)
{
    int *i;

    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case NUMBER:
            i = (int *) malloc(sizeof(int));
            assert(i);
            *i = node_get_number(node);
            g_hash_table_add(elems, i);
            break;
        default:
            find_ints_in_cons(node_get_left(node), elems);
            find_ints_in_cons(node_get_right(node), elems);
            break;
    }
}

void translate_initial_term(node_ptr node, GArray *vars,
        node_ptr fax, node_ptr fay)
{
    GArray *indexes = g_array_new(FALSE, FALSE, sizeof(node_ptr));

    node_ptr x = NULL;
    node_ptr y = NULL;

    if (fax == Nil) {
        find_different_indexes_in_term(node, indexes);

        if (indexes->len > 2) {
            printf("\n");
            fprintf(stderr, "initial formula: "
                    "too many indexes in a term\n");
            exit(1);
        }

        if (indexes->len > 0) {
            printf(" (=>");
            if (indexes->len == 1) {
                x = g_array_index(indexes, node_ptr, 0);
                printf(" (= x ");
                translate_initial_index(x, vars);
                printf(")");
            } else {
                x = g_array_index(indexes, node_ptr, 0);
                y = g_array_index(indexes, node_ptr, 1);
                printf(" (and");
                printf(" (= x ");
                translate_initial_index(x, vars);
                printf(")");
                printf(" (= y ");
                translate_initial_index(y, vars);
                printf(")");
                printf(")");
            }
        }

        translate_initial_term_replacing_indexes(node, vars, x, y);


        if (indexes->len > 0)
            printf(")");

        g_array_free(indexes, TRUE);
    } else {
        translate_initial_term_replacing_indexes(node, vars, fax, fay);
    }
}

void find_different_indexes_in_term(node_ptr node, GArray *indexes)
{
    node_ptr idx, n;
    int i;

    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case INDEXED:
            idx = node_get_left(node_get_right(node));
            for (i = 0; i < indexes->len; i++) {
                n = g_array_index(indexes, node_ptr, i);
                if (indexes_equal(n, idx))
                    break;
            }
            if (i == indexes->len)
                g_array_append_val(indexes, idx);
            break;
        case ID:
            // fall through
        case NUMBER:
            // fall through
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            break;
        default:
            find_different_indexes_in_term(node_get_left(node), indexes);
            find_different_indexes_in_term(node_get_right(node), indexes);
    }
}

int indexes_equal(node_ptr i1, node_ptr i2)
{
    if (i1 == NULL && i2 == NULL)
        return 1;
    if (i1 == NULL || i2 == NULL)
        return 0;

    if (node_get_type(i1) == ID && node_get_type(i2) == ID)
        return !strcmp((char *) node_get_left(i1), (char *) node_get_left(i2));
    else if (node_get_type(i1) == NUMBER && node_get_type(i2) == NUMBER)
        return (node_get_number(i1) == node_get_number(i2));
    else // Syntactically compare expressions
        return (node_get_type(i1) == node_get_type(i2) &&
                indexes_equal(node_get_left(i1), node_get_left(i2)) &&
                indexes_equal(node_get_right(i1), node_get_right(i2)));
}

void translate_id(node_ptr node, GArray *vars, node_ptr repl_x, node_ptr repl_y,
        GString *out)
{
    int i;
    struct mcmt_var v;

    if (repl_x && !strcmp((char *) node_get_left(node),
                (char *) node_get_left(repl_x))) {
        printf_redir(out, "x");
    } else if (repl_y && !strcmp((char *) node_get_left(node),
                (char *) node_get_left(repl_y))) {
        printf_redir(out, "y");
    } else {
        i = get_index_of_var(vars, (char *) node_get_left(node));
        if (i == -1) {
            printf("\n");
            fprintf(stderr, "cannot translate identifier %s\n",
                    (char *) node_get_left(node));
            exit(1);
        }
        v = g_array_index(vars, struct mcmt_var, i);
        printf_redir(out, "%s", v.mcmt_name);
    }
}

void translate_initial_index(node_ptr node, GArray *vars)
{
    if (node_get_type(node) == ID)
        translate_id(node, vars, NULL, NULL, NULL);
    else if (node_get_type(node) == NUMBER)
        printf("%d", node_get_number(node));
    else
        translate_initial_term(node, vars, Nil, Nil);
}

void translate_size(node_ptr node, GArray *vars, GString *out)
{
    int i;
    struct mcmt_var v;

    printf_redir(out, " ");
    i = get_index_of_var(vars, (char *) node_get_left(node_get_left(node)));
    if (i == -1) {
        printf("\n");
        fprintf(stderr, "cannot translate identifier %s\n",
                (char *) node_get_left(node_get_left(node)));
        exit(1);
    }
    v = g_array_index(vars, struct mcmt_var, i);
    v = g_array_index(vars, struct mcmt_var, v.size_var);
    printf_redir(out, "%s", v.mcmt_name);
}

void translate_initial_term_replacing_indexes(node_ptr node, GArray *vars,
        node_ptr repl_x, node_ptr repl_y)
{
    node_ptr id, idx;

    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case LE:
            printf(" (<=");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf(")");
            break;
        case GE:
            printf(" (>=");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf(")");
            break;
        case LT:
            printf(" (<");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf(")");
            break;
        case GT:
            printf(" (>");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf(")");
            break;
        case EQUAL:
            printf(" (=");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf(")");
            break;
        case NOTEQUAL:
            printf(" (not (=");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf("))");
            break;
        case PLUS:
            printf(" (+");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf(")");
            break;
        case MINUS:
            printf(" (+");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            printf(" (* -1");
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf("))");
            break;
        case TIMES:
            printf(" (*");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf(")");
            break;
        case FRAC:
            printf(" (div");
            translate_initial_term_replacing_indexes(node_get_left(node),
                    vars, repl_x, repl_y);
            translate_initial_term_replacing_indexes(node_get_right(node),
                    vars, repl_x, repl_y);
            printf(")");
            break;
        case SIZE:
            translate_size(node, vars, NULL);
            break;
        case ID:
            printf(" ");
            translate_id(node, vars, repl_x, repl_y, NULL);
            break;
        case NUMBER:
            printf(" %d", node_get_number(node));
            break;
        case BOOL_TRUE:
            printf(" true");
            break;
        case BOOL_FALSE:
            printf(" false");
            break;
        case INDEXED:
            id = node_get_left(node);
            idx = node_get_left(node_get_right(node));
            printf(" ");
            translate_id(id, vars, NULL, NULL, NULL);
            printf("[");
            if (!repl_x)
                translate_initial_index(idx, vars);
            else if (indexes_equal(idx, repl_x))
                printf("x");
            else if (repl_y && indexes_equal(idx, repl_y))
                printf("y");
            else
                translate_initial_index(idx, vars);
            printf("]");
            break;
        default:
            break;
    }
}

void output_unsafe_formula(GArray *vars)
{
    int i;
    struct mcmt_var v;

    printf("\n");
    for (i = 0; i < vars->len; i++) {
        v = g_array_index(vars, struct mcmt_var, i);
        if (v.kind != KIND_TOKEN && v.kind != KIND_TOKEN_SET)
            continue;
        if (!v.is_dst)
            continue;

        if (v.kind == KIND_TOKEN)
            printf(":u_cnj (= %s true)\n", v.mcmt_name);
        if (v.kind == KIND_TOKEN_SET)
            printf(":u_cnj (= %s[z1] true)\n", v.mcmt_name);
    }
}

void output_communication_transitions(node_ptr node, GArray *vars)
{
    int i;
    node_ptr n;

    GArray *conns = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    find_connections(node, conns);

    for (i = 0; i < conns->len; i++) {
        n = g_array_index(conns, node_ptr, i);
        translate_connection(n, node, vars);
    }

    g_array_free(conns, TRUE);
}

void find_connections(node_ptr node, GArray *conns)
{
    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case CONNECTION:
            g_array_append_val(conns, node);
            break;
        case ID:
            // fall through
        case NUMBER:
            // fall through
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            break;
        default:
            find_connections(node_get_right(node), conns);
            find_connections(node_get_left(node), conns);
            break;
    }
}

void translate_connection(node_ptr node, node_ptr tree, GArray *vars)
{
    int c;

    c = determine_connection_case(node);
    conn_count++;

    switch(c) {
        case CONN_CASE_1:
            translate_connection_case_1(node, tree, vars);
            break;
        case CONN_CASE_2:
            translate_connection_case_2(node, tree, vars);
            break;
        case CONN_CASE_3:
            translate_connection_case_3(node, tree, vars);
            break;
        case CONN_CASE_4a:
            translate_connection_case_4a(node, tree, vars);
            break;
        case CONN_CASE_4b:
            translate_connection_case_4b(node, tree, vars);
            break;
        case CONN_CASE_5:
            translate_connection_case_5(node, tree, vars);
            break;
        case CONN_CASE_6:
            translate_connection_case_6(node, tree, vars);
            break;
        case CONN_CASE_7:
            translate_connection_case_7(node, tree, vars);
            break;
        case CONN_CASE_8:
            translate_connection_case_8(node, tree, vars);
            break;
        case CONN_CASE_9:
            translate_connection_case_9(node, tree, vars);
            break;
        default:
            printf("\n");
            fprintf(stderr, "invalid connection case\n");
            exit(1);
            break;
    }
}

int determine_connection_case(node_ptr node)
{
    node_ptr s, d;
    int s_single = 0, s_var = 0, s_const = 0;
    int d_single = 0, d_var = 0, d_const = 0;
    int vars_equal = 0;

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    if (node_get_type(s) == ID)
        s_single = 1;
    else if (node_get_type(s) == INDEXED && 
            node_get_type(node_get_left(node_get_right(s))) == ID)
        s_var = 1;
    else if (node_get_type(s) == INDEXED && 
            node_get_type(node_get_left(node_get_right(s))) == NUMBER)
        s_const = 1;

    if (node_get_type(d) == ID)
        d_single = 1;
    else if (node_get_type(d) == INDEXED && 
            node_get_type(node_get_left(node_get_right(d))) == ID)
        d_var = 1;
    else if (node_get_type(d) == INDEXED && 
            node_get_type(node_get_left(node_get_right(d))) == NUMBER)
        d_const = 1;
    else {
        printf("\n");
        fprintf(stderr, "invalid subcomponent specification in connection\n");
        exit(1);
    }

    if (s_var && d_var)
        vars_equal = !strcmp(
                (char *) node_get_left(node_get_left(node_get_right(s))),
                (char *) node_get_left(node_get_left(node_get_right(d))));

    if (s_const && d_const)
        return CONN_CASE_1;
    else if (s_const && d_var)
        return CONN_CASE_2;
    else if (s_var && d_const)
        return CONN_CASE_3;
    else if (s_var && d_var && !vars_equal)
        return CONN_CASE_4a;
    else if (s_var && d_var && vars_equal)
        return CONN_CASE_4b;
    else if (s_single && d_var)
        return CONN_CASE_5;
    else if (s_single && d_const)
        return CONN_CASE_6;
    else if (s_var && d_single)
        return CONN_CASE_7;
    else if (s_const && d_single)
        return CONN_CASE_8;
    else if (s_single && d_single)
        return CONN_CASE_9;

    printf("\n");
    fprintf(stderr, "unrecognized connection case\n");
    exit(1);
}

void translate_connection_case_1(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr s_idx, d_idx;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    struct mcmt_var v, s_index_set, d_index_set;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = node_get_left(s);
    d_id = node_get_left(d);

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    s_idx = node_get_left(node_get_right(s));
    d_idx = node_get_left(node_get_right(d));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, guard_atoms);
    find_conjuncts(d_guard, guard_atoms);

    i = get_index_of_var(vars, (char *) node_get_left(s_id));
    v = g_array_index(vars, struct mcmt_var, i);
    s_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    i = get_index_of_var(vars, (char *) node_get_left(d_id));
    v = g_array_index(vars, struct mcmt_var, i);
    d_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 1\n", conn_count);
    printf(":transition\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s[", s_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", s_token_mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", d_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("] true)");
    printf("\n");

    printf(":numcases 2\n");

    printf(":case (= j ");
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf(")\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j ");
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("))\n");
    output_upds_set_one_to_true(vars, NULL);

    g_array_free(guard_atoms, TRUE);
}

void translate_connection_case_2(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr s_idx, d_idx;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    struct mcmt_var v, s_index_set, d_index_set;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = node_get_left(s);
    d_id = node_get_left(d);

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    s_idx = node_get_left(node_get_right(s));
    d_idx = node_get_left(node_get_right(d));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, comp_guard_atoms);
    find_conjuncts(d_guard, comp_guard_atoms);

    i = get_index_of_var(vars, (char *) node_get_left(s_id));
    v = g_array_index(vars, struct mcmt_var, i);
    s_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    i = get_index_of_var(vars, (char *) node_get_left(d_id));
    v = g_array_index(vars, struct mcmt_var, i);
    d_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 2\n", conn_count);
    printf(":transition\n");
    printf(":var x\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, d_idx, "x", NULL, NULL,
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s[", s_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, d_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", s_token_mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, d_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", d_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(d_idx, vars, d_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");
    printf("\n");

    printf(":numcases 2\n");

    printf(":case (= j x)\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j x))\n");
    output_upds_set_one_to_true(vars, NULL);

    g_array_free(guard_atoms, TRUE);
    g_array_free(comp_guard_atoms, TRUE);
}

void translate_connection_case_3(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr s_idx, d_idx;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    struct mcmt_var v, s_index_set, d_index_set;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = node_get_left(s);
    d_id = node_get_left(d);

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    s_idx = node_get_left(node_get_right(s));
    d_idx = node_get_left(node_get_right(d));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, comp_guard_atoms);
    find_conjuncts(d_guard, comp_guard_atoms);

    i = get_index_of_var(vars, (char *) node_get_left(s_id));
    v = g_array_index(vars, struct mcmt_var, i);
    s_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    i = get_index_of_var(vars, (char *) node_get_left(d_id));
    v = g_array_index(vars, struct mcmt_var, i);
    d_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    // first transition
    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 3, first transition\n", conn_count);
    printf(":transition\n");
    printf(":var x\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, s_idx, "x", NULL, NULL,
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s[", s_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", s_token_mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", d_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(d_idx, vars, s_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= x ");
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf(")");

    printf("\n");

    printf(":numcases 2\n");

    printf(":case (= j x)\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j x))\n");
    output_upds_set_one_to_true(vars, NULL);

    // second transition
    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 3, second transition\n", conn_count);
    printf(":transition\n");
    printf(":var x\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, s_idx, "x", NULL, NULL,
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s[", s_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", s_token_mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", d_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(d_idx, vars, s_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (not (= x ");
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("))");

    printf("\n");

    printf(":numcases 3\n");

    printf(":case (= j x)\n");
    output_upds_set_one_to_true(vars, NULL);

    printf(":case (not (= j x)) (= j ");
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf(")\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j x)) (not (= j ");
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("))\n");
    output_upds_set_one_to_true(vars, NULL);

    g_array_free(guard_atoms, TRUE);
    g_array_free(comp_guard_atoms, TRUE);
}

void translate_connection_case_4a(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr s_idx, d_idx;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    struct mcmt_var v, s_index_set, d_index_set;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = node_get_left(s);
    d_id = node_get_left(d);

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    s_idx = node_get_left(node_get_right(s));
    d_idx = node_get_left(node_get_right(d));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, comp_guard_atoms);
    find_conjuncts(d_guard, comp_guard_atoms);

    i = get_index_of_var(vars, (char *) node_get_left(s_id));
    v = g_array_index(vars, struct mcmt_var, i);
    s_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    i = get_index_of_var(vars, (char *) node_get_left(d_id));
    v = g_array_index(vars, struct mcmt_var, i);
    d_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    // first transition
    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 4a, first transition\n", conn_count);
    printf(":transition\n");
    printf(":var x\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, s_idx, "x", d_idx, "x",
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s[", s_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", d_idx, "x",
            NULL);
    printf("] true)");

    printf(" (= %s[", s_token_mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", d_idx, "x",
            NULL);
    printf("] true)");

    printf(" (= %s[", d_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(d_idx, vars, s_idx, "x", d_idx, "x",
            NULL);
    printf("] true)");

    printf("\n");

    printf(":numcases 2\n");

    printf(":case (= j x)\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j x))\n");
    output_upds_set_one_to_true(vars, NULL);

    // second transition
    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 4a, second transition\n", conn_count);
    printf(":transition\n");
    printf(":var x\n");
    printf(":var y\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, s_idx, "x", d_idx, "y",
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s[", s_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", d_idx, "y",
            NULL);
    printf("] true)");

    printf(" (= %s[", s_token_mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", d_idx, "y",
            NULL);
    printf("] true)");

    printf(" (= %s[", d_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(d_idx, vars, s_idx, "x", d_idx, "y",
            NULL);
    printf("] true)");

    printf("\n");

    printf(":numcases 3\n");

    printf(":case (= j x)\n");
    output_upds_set_one_to_true(vars, NULL);

    printf(":case (not (= j x)) (= j y)\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j x)) (not (= j y))\n");
    output_upds_set_one_to_true(vars, NULL);

    g_array_free(guard_atoms, TRUE);
    g_array_free(comp_guard_atoms, TRUE);
}

void translate_connection_case_4b(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr s_idx, d_idx;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    struct mcmt_var v, s_index_set, d_index_set;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = node_get_left(s);
    d_id = node_get_left(d);

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    s_idx = node_get_left(node_get_right(s));
    d_idx = node_get_left(node_get_right(d));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, comp_guard_atoms);
    find_conjuncts(d_guard, comp_guard_atoms);

    i = get_index_of_var(vars, (char *) node_get_left(s_id));
    v = g_array_index(vars, struct mcmt_var, i);
    s_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    i = get_index_of_var(vars, (char *) node_get_left(d_id));
    v = g_array_index(vars, struct mcmt_var, i);
    d_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 4b\n", conn_count);
    printf(":transition\n");
    printf(":var x\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, s_idx, "x", d_idx, "x",
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s[", s_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", d_idx, "x",
            NULL);
    printf("] true)");

    printf(" (= %s[", s_token_mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", d_idx, "x",
            NULL);
    printf("] true)");

    printf(" (= %s[", d_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(d_idx, vars, s_idx, "x", d_idx, "x",
            NULL);
    printf("] true)");

    printf("\n");

    printf(":numcases 2\n");

    printf(":case (= j x)\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j x))\n");
    output_upds_set_one_to_true(vars, NULL);

    g_array_free(guard_atoms, TRUE);
    g_array_free(comp_guard_atoms, TRUE);
}

void translate_connection_case_5(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr d_idx;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    struct mcmt_var v, d_index_set;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = s;
    d_id = node_get_left(d);

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    d_idx = node_get_left(node_get_right(d));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, comp_guard_atoms);
    find_conjuncts(d_guard, comp_guard_atoms);

    i = get_index_of_var(vars, (char *) node_get_left(d_id));
    v = g_array_index(vars, struct mcmt_var, i);
    d_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 5\n", conn_count);
    printf(":transition\n");
    printf(":var x\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, d_idx, "x", NULL, NULL,
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s true)", s_token_mcmt_name);

    printf(" (= %s[", d_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(d_idx, vars, d_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf("\n");

    printf(":numcases 2\n");

    printf(":case (= j x)\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j x))\n");
    output_upds_set_one_to_true(vars, NULL);

    g_array_free(guard_atoms, TRUE);
    g_array_free(comp_guard_atoms, TRUE);
}

void translate_connection_case_6(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr d_idx;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    struct mcmt_var v, d_index_set;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = s;
    d_id = node_get_left(d);

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    d_idx = node_get_left(node_get_right(d));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, comp_guard_atoms);
    find_conjuncts(d_guard, comp_guard_atoms);

    i = get_index_of_var(vars, (char *) node_get_left(d_id));
    v = g_array_index(vars, struct mcmt_var, i);
    d_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 6\n", conn_count);
    printf(":transition\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s true)", s_token_mcmt_name);

    printf(" (= %s[", d_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("] true)");

    printf("\n");

    printf(":numcases 2\n");

    printf(":case (= j ");
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf(")\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j ");
    translate_term_replacing_ids_and_nums(d_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("))\n");
    output_upds_set_one_to_true(vars, NULL);

    g_array_free(guard_atoms, TRUE);
    g_array_free(comp_guard_atoms, TRUE);
}

void translate_connection_case_7(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr s_idx;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    struct mcmt_var v, s_index_set;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = node_get_left(s);
    d_id = d;

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    s_idx = node_get_left(node_get_right(s));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, comp_guard_atoms);
    find_conjuncts(d_guard, comp_guard_atoms);

    i = get_index_of_var(vars, (char *) node_get_left(s_id));
    v = g_array_index(vars, struct mcmt_var, i);
    s_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 7\n", conn_count);
    printf(":transition\n");
    printf(":var x\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, s_idx, "x", NULL, NULL,
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s[", s_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", s_token_mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, s_idx, "x", NULL, NULL,
            NULL);
    printf("] true)");

    printf("\n");

    printf(":numcases 2\n");

    printf(":case (= j x)\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    printf(":case (not (= j x))\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    g_array_free(guard_atoms, TRUE);
    g_array_free(comp_guard_atoms, TRUE);
}

void translate_connection_case_8(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr s_idx;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    struct mcmt_var v, s_index_set;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = node_get_left(s);
    d_id = d;

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    s_idx = node_get_left(node_get_right(s));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, comp_guard_atoms);
    find_conjuncts(d_guard, comp_guard_atoms);

    i = get_index_of_var(vars, (char *) node_get_left(s_id));
    v = g_array_index(vars, struct mcmt_var, i);
    s_index_set = g_array_index(vars, struct mcmt_var, v.index_set);

    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 8\n", conn_count);
    printf(":transition\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s[", s_index_set.mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("] true)");

    printf(" (= %s[", s_token_mcmt_name);
    translate_term_replacing_ids_and_nums(s_idx, vars, NULL, NULL, NULL, NULL,
            NULL);
    printf("] true)");

    printf("\n");

    printf(":numcases 1\n");

    printf(":case true\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    g_array_free(guard_atoms, TRUE);
    g_array_free(comp_guard_atoms, TRUE);
}

void translate_connection_case_9(node_ptr node, node_ptr tree, GArray *vars)
{
    node_ptr s, d;
    node_ptr s_id, d_id;
    node_ptr guard, s_guard, d_guard;
    GArray *guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guard_atoms = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    char s_token_mcmt_name[MAX_ID_LEN + 10];
    char d_token_mcmt_name[MAX_ID_LEN + 10];

    s = node_get_left(node_get_left(node));
    d = node_get_right(node_get_left(node));

    s_id = s;
    d_id = d;

    sprintf(s_token_mcmt_name, "_%s_token", (char *) node_get_left(s_id));
    sprintf(d_token_mcmt_name, "_%s_token", (char *) node_get_left(d_id));

    guard = node_get_left(node_get_left(node_get_right(node)));
    s_guard = find_subcomp_guard(tree, (char *) node_get_left(s_id));
    d_guard = find_subcomp_guard(tree, (char *) node_get_left(d_id));

    if (s_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(s_id));
        exit(1);
    }
    if (d_guard == NULL) {
        printf("\n");
        fprintf(stderr, "cannot find guard  for component %s\n",
                (char *) node_get_left(d_id));
        exit(1);
    }

    find_conjuncts(guard, guard_atoms);
    find_conjuncts(s_guard, comp_guard_atoms);
    find_conjuncts(d_guard, comp_guard_atoms);

    printf("\n");
    printf(":comment transition t%d\n", ++mcmt_trans_count);
    printf(":comment connection %d, case 9\n", conn_count);
    printf(":transition\n");
    printf(":var j\n");

    printf(":guard true");
    for (i = 0; i < guard_atoms->len; i++) {
        n = g_array_index(guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }
    for (i = 0; i < comp_guard_atoms->len; i++) {
        n = g_array_index(comp_guard_atoms, node_ptr, i);
        printf(" ");
        translate_atom_replacing_ids_and_nums(n, vars, NULL, NULL, NULL, NULL,
                NULL);
    }

    printf(" (= %s true)", s_token_mcmt_name);

    printf("\n");

    printf(":numcases 1\n");

    printf(":case true\n");
    output_upds_set_one_to_true(vars, d_token_mcmt_name);

    g_array_free(guard_atoms, TRUE);
    g_array_free(comp_guard_atoms, TRUE);
}

void find_conjuncts(node_ptr node, GArray *conjs)
{
    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case AND:
            find_conjuncts(node_get_left(node), conjs);
            find_conjuncts(node_get_right(node), conjs);
            break;
        default:
            g_array_append_val(conjs, node);
            break;
    }
}

void translate_atom_replacing_ids_and_nums(node_ptr node, GArray *vars,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y, GString *out)
{
    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case NOT:
            translate_in_replacing_ids_and_nums(node_get_left(node), vars, 0,
                    repl_x, x, repl_y, y, out);
            break;
        case IN:
            translate_in_replacing_ids_and_nums(node, vars, 1, repl_x, x,
                    repl_y, y, out);
            break;
        case EQUAL:
            if (is_index_set_spec(node, vars)) {
                printf("\n");
                fprintf(stderr, "index set specification not allowed here\n");
                exit(1);
            } else {
                translate_term_replacing_ids_and_nums(node, vars, repl_x, x,
                        repl_y, y, out);
            }
            break;
        case LE:
            // fall through
        case GE:
            // fall through
        case LT:
            // fall through
        case GT:
            // fall through
        case NOTEQUAL:
            translate_term_replacing_ids_and_nums(node, vars, repl_x, x,
                    repl_y, y, out);
            break;
        case ID:
            // fall through
        case NUMBER:
            assert(0);
            break;
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            translate_term_replacing_ids_and_nums(node, vars, repl_x, x,
                    repl_y, y, out);
            break;
        default:
            break;
    }
}

void translate_in_replacing_ids_and_nums(node_ptr node, GArray *vars, int val,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y, GString *out)
{
    node_ptr term;
    node_ptr set;
    int i;
    struct mcmt_var v;

    term = node_get_left(node);
    set = node_get_right(node);
    i = get_index_of_var(vars, (char *) node_get_left(set));
    v = g_array_index(vars, struct mcmt_var, i);

    printf_redir(out, "(= %s[", v.mcmt_name);
    translate_term_replacing_ids_and_nums(term, vars, repl_x, x, repl_y, y,
            out);
    if (val)
        printf_redir(out, "] true)");
    else
        printf_redir(out, "] false)");
}

void translate_term_replacing_ids_and_nums(node_ptr node, GArray *vars,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y, GString *out)
{
    node_ptr id, idx;
    int num;

    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case LE:
            printf_redir(out, "(<= ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, ")");
            break;
        case GE:
            printf_redir(out, "(>= ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, ")");
            break;
        case LT:
            printf_redir(out, "(< ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, ")");
            break;
        case GT:
            printf_redir(out, "(> ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, ")");
            break;
        case EQUAL:
            printf_redir(out, "(= ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, ")");
            break;
        case NOTEQUAL:
            printf_redir(out, "(not (=");
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, "))");
            break;
        case PLUS:
            printf_redir(out, "(+ ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, ")");
            break;
        case MINUS:
            printf_redir(out, "(+ ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " (* -1 ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, "))");
            break;
        case TIMES:
            printf_redir(out, "(* ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, ")");
            break;
        case FRAC:
            printf_redir(out, "(div ");
            translate_term_replacing_ids_and_nums(node_get_left(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, " ");
            translate_term_replacing_ids_and_nums(node_get_right(node), vars,
                    repl_x, x, repl_y, y, out);
            printf_redir(out, ")");
            break;
        case SIZE:
            translate_size(node, vars, out);
            break;
        case ID:
            if (repl_x && node_get_type(repl_x) == ID &&
                    !strcmp((char *) node_get_left(node),
                            (char *) node_get_left(repl_x)))
                printf_redir(out, "%s", x);
            else if (repl_y && node_get_type(repl_y) == ID &&
                    !strcmp((char *) node_get_left(node),
                            (char *) node_get_left(repl_y)))
                printf_redir(out, "%s", y);
            else
                translate_id(node, vars, NULL, NULL, out);
            break;
        case NUMBER:
            num = node_get_number(node);
            if (repl_x && node_get_type(repl_x) == NUMBER &&
                    num == node_get_number(repl_x))
                printf_redir(out, "%s", x);
            else if (repl_y && node_get_type(repl_y) == NUMBER &&
                    num == node_get_number(repl_y))
                printf_redir(out, "%s", y);
            else
                printf_redir(out, "%d", num);
            break;
        case BOOL_TRUE:
            printf_redir(out, "true");
            break;
        case BOOL_FALSE:
            printf_redir(out, "false");
            break;
        case INDEXED:
            id = node_get_left(node);
            idx = node_get_left(node_get_right(node));
            translate_id(id, vars, NULL, NULL, out);
            printf_redir(out, "[");
            translate_term_replacing_ids_and_nums(idx, vars, repl_x, x,
                    repl_y, y, out);
            printf_redir(out, "]");
            break;
        default:
            break;
    }
}

node_ptr find_subcomp_guard(node_ptr node, char *name)
{
    node_ptr result;
    char *id;

    if (node == NULL)
        return NULL;

    switch(node_get_type(node)) {
        case SUBCOMPONENT:
            id = (char *) node_get_left(node_get_left(node_get_left(node)));
            if (!strcmp(id, name))
                result = node_get_left(node_get_right(node));
            else
                result = NULL;
            break;
        case ID:
            // fall through
        case NUMBER:
            // fall through
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            result = NULL;
            break;
        default:
            result = find_subcomp_guard(node_get_left(node), name);
            if (!result)
                result = find_subcomp_guard(node_get_right(node), name);
    }

    return result;
}

void output_upds_set_one_to_true(GArray *vars, char *mcmt_name)
{
    int i;
    struct mcmt_var v;

    for (i = 0; i < vars->len; i++) {
        v = g_array_index(vars, struct mcmt_var, i);
        printf(":val ");
        if (mcmt_name && !strcmp(v.mcmt_name, mcmt_name)) {
            printf("true\n");
        } else {
            if (v.kind == KIND_INDEX_SET || v.kind == KIND_PARAM_SET
                    || v.kind == KIND_TOKEN_SET)
                printf("%s[j]\n", v.mcmt_name);
            else
                printf("%s\n", v.mcmt_name);
        }
    }
}

void output_reconfiguration_transitions(node_ptr node, GArray *vars)
{
    int i;
    node_ptr n;

    GArray *reconfs = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    find_reconfigurations(node, reconfs, 0);

    for (i = 0; i < reconfs->len; i++) {
        n = g_array_index(reconfs, node_ptr, i);
        translate_reconfiguration(n, node, vars);
    }

    g_array_free(reconfs, TRUE);
}

void find_reconfigurations(node_ptr node, GArray *reconfs, int in_reconf)
{
    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case RECONFIGURATIONS:
            find_reconfigurations(node_get_left(node), reconfs, 1);
            break;
        case CONS:
            find_reconfigurations(node_get_right(node), reconfs, in_reconf);
            find_reconfigurations(node_get_left(node), reconfs, in_reconf);
            break;
        case ID:
            // fall through
        case NUMBER:
            // fall through
            break;
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            if (in_reconf)
                g_array_append_val(reconfs, node);
            break;
        default:
            if (in_reconf) {
                g_array_append_val(reconfs, node);
            } else {
                find_reconfigurations(node_get_right(node), reconfs, in_reconf);
                find_reconfigurations(node_get_left(node), reconfs, in_reconf);
            }
            break;
    }
}

void translate_reconfiguration(node_ptr node, node_ptr tree, GArray *vars)
{
    int i, j, len;
    node_ptr n, nn, nf, n1, n2;
    GArray *invars = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *comp_guards = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GArray *arr_indices = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    GSList *indices = NULL, *e, *ei, *ej;
    struct case_index *cip;
    char *var_x = NULL, *set_x = NULL, *var_y = NULL, *set_y = NULL;
    int set_x_compl, set_y_compl;

    if (node_get_type(node) == EXISTS) {
        var_x = (char *) node_get_left(node_get_left(node_get_left(node)));
        set_x = (char *) node_get_left(node_get_right(node_get_left(node)));
        set_x_compl = (node_get_type(node_get_left(node)) == SCOPE_COMPLEMENT);
        node = node_get_right(node);
    }
    if (node_get_type(node) == EXISTS) {
        var_y = (char *) node_get_left(node_get_left(node_get_left(node)));
        set_y = (char *) node_get_left(node_get_right(node_get_left(node)));
        set_y_compl = (node_get_type(node_get_left(node)) == SCOPE_COMPLEMENT);
        node = node_get_right(node);
    }

    find_invariant_formulas(tree, invars, 0);
    for (i = 0; i < invars->len; i++) {
        n = g_array_index(invars, node_ptr, i);
        find_reconf_different_indexes(n, vars, arr_indices, 0);
    }
    find_component_guards(tree, comp_guards);
    for (i = 0; i < comp_guards->len; i++) {
        n = g_array_index(comp_guards, node_ptr, i);
        find_reconf_different_indexes(n, vars, arr_indices, 0);
    }
    find_reconf_different_indexes(node, vars, arr_indices, 0);
    for (i = 0; i < arr_indices->len; i++) {
        n = g_array_index(arr_indices, node_ptr, i);

        if (node_get_type(n) == NUMBER) {
            cip = (struct case_index *) malloc(sizeof(struct case_index));
            assert(cip);
            cip->type = INDEX_CONST;
            cip->constant = node_get_number(n);
            indices = g_slist_insert_sorted(indices, cip, compare_case_indexes);
        } else if (node_get_type(n) == ID) {
            if ((!var_x || strcmp((char *) node_get_left(n), var_x))
                && (!var_y || strcmp((char *) node_get_left(n), var_y))) {
                printf("\n");
                fprintf(stderr, "unbound index variable or"
                        " illegal parameter in index\n");
                exit(1);
            }
        } else {
            printf("\n");
            fprintf(stderr, "complex index variable\n");
            exit(1);
        }
    }
    cip = (struct case_index *) malloc(sizeof(struct case_index));
    assert(cip);
    cip->type = INDEX_OTHER;
    indices = g_slist_insert_sorted(indices, cip, compare_case_indexes);

    reconf_count++;
    if (!var_x) {
        // Case 1: only constants
        printf("\n");
        printf(":comment transition t%d\n", ++mcmt_trans_count);
        printf(":comment reconfiguration %d, case 1\n", reconf_count);
        translate_simple_reconfiguration(node, tree, vars, indices);
    } else if (!var_y) {
        // Case 2: one existentially quantified variable and constants

        // first n transitions
        len = g_slist_length(indices) - 1; // last one is INDEX_OTHER
        for (e = indices, i = 0; i < len; e = g_slist_next(e), i++) {
            n = new_number_node(((struct case_index *) e->data)->constant);
            n1 = new_node(IN, n, new_node(ID, (node_ptr) set_x, Nil));
            if (set_x_compl)
                n1 = new_node(NOT, n1, Nil);
            nf = rewrite_formula_replacing_var(node, var_x, n);
            nf = new_node(AND, n1, nf);

            //assert(node_get_type(n) == NUMBER);
            printf("\n");
            printf(":comment transition t%d\n", ++mcmt_trans_count);
            printf(":comment reconfiguration %d, case 2,"
                    " transition %d of %d, %s -> %d\n", reconf_count,
                    i + 1, len + 1, var_x, node_get_number(n));
            translate_simple_reconfiguration(nf, tree, vars, indices);
        }

        // last transition
        cip = (struct case_index *) malloc(sizeof(struct case_index));
        assert(cip);
        cip->type = INDEX_VAR_X;
        strcpy(cip->orig_name, var_x);
        indices = g_slist_insert_sorted(indices, cip, compare_case_indexes);

        n1 = new_node(IN,
                      new_node(ID, (node_ptr) var_x, Nil),
                      new_node(ID, (node_ptr) set_x, Nil));
        if (set_x_compl)
            n1 = new_node(NOT, n1, Nil);
        nf = new_node(AND, n1, node);

        for (e = indices; e; e = g_slist_next(e)) {
            if (((struct case_index *)e->data)->type == INDEX_CONST) {
                n1 = new_node(NOTEQUAL,
                              new_node(ID, (node_ptr) var_x, Nil),
                              new_number_node(((struct case_index *)
                                               e->data)->constant));
                nf = new_node(AND, n1, nf);
            }
        }

        printf("\n");
        printf(":comment transition t%d\n", ++mcmt_trans_count);
        printf(":comment reconfiguration %d, case 2,"
                " transition %d of %d, %s -> x\n", reconf_count, len + 1,
                len + 1, var_x);
        translate_simple_reconfiguration(nf, tree, vars, indices);
    } else {
        // Case 3: two existentially quantified variables and constants

        // n^2 transitions
        // x and y equal to some constants
        len = g_slist_length(indices) - 1; // last one is INDEX_OTHER
        for (ei = indices, i = 0; i < len; ei = g_slist_next(ei), i++) {
            n = new_number_node(((struct case_index *) ei->data)->constant);
            for (ej = indices, j = 0; j < len; ej = g_slist_next(ej), j++) {
                nn = new_number_node(((struct case_index *)ej->data)->constant);
                n1 = new_node(IN, n, new_node(ID, (node_ptr) set_x, Nil));
                if (set_x_compl)
                    n1 = new_node(NOT, n1, Nil);
                n2 = new_node(IN, nn, new_node(ID, (node_ptr) set_y, Nil));
                if (set_y_compl)
                    n2 = new_node(NOT, n2, Nil);
                nf = rewrite_formula_replacing_var(node, var_x, n);
                nf = rewrite_formula_replacing_var(nf, var_y, nn);
                nf = new_node(AND, n2, nf);
                nf = new_node(AND, n1, nf);

                printf("\n");
                printf(":comment transition t%d\n", ++mcmt_trans_count);
                printf(":comment reconfiguration %d, case 3,"
                       " transition %d of %d, %s -> %d, %s -> %d\n",
                       reconf_count, i * len + j + 1, len * len + 2 * len + 2,
                       var_x, node_get_number(n), var_y, node_get_number(nn));
                translate_simple_reconfiguration(nf, tree, vars, indices);
            }
        }

        // n transitions
        // x different from all constants, y equal to some constant
        len = g_slist_length(indices) - 1; // last one is INDEX_OTHER
        cip = (struct case_index *) malloc(sizeof(struct case_index));
        assert(cip);
        cip->type = INDEX_VAR_X;
        strcpy(cip->orig_name, var_x);
        indices = g_slist_insert_sorted(indices, cip, compare_case_indexes);
        for (e = indices, i = 0; e; e = g_slist_next(e)) {
            if (((struct case_index *)e->data)->type != INDEX_CONST)
                continue;
            n1 = new_node(IN,
                    new_node(ID, (node_ptr) var_x, Nil),
                    new_node(ID, (node_ptr) set_x, Nil));
            if (set_x_compl)
                n1 = new_node(NOT, n1, Nil);

            n = new_number_node(((struct case_index *) e->data)->constant);
            n2 = new_node(IN, n, new_node(ID, (node_ptr) set_y, Nil));
            if (set_y_compl)
                n2 = new_node(NOT, n2, Nil);
            nf = rewrite_formula_replacing_var(node, var_y, n);
            nf = new_node(AND, n1, nf);
            nf = new_node(AND, n2, nf);

            for (ej = indices; ej; ej = g_slist_next(ej)) {
                if (((struct case_index *)ej->data)->type == INDEX_CONST) {
                    n1 = new_node(NOTEQUAL,
                            new_node(ID, (node_ptr) var_x, Nil),
                            new_number_node(((struct case_index *)
                                    ej->data)->constant));
                    nf = new_node(AND, n1, nf);
                }
            }

            printf("\n");
            printf(":comment transition t%d\n", ++mcmt_trans_count);
            printf(":comment reconfiguration %d, case 3,"
                    " transition %d of %d, %s -> x, %s -> %d\n", reconf_count,
                    len * len + i + 1, len * len + 2 * len + 2, var_x, var_y,
                    node_get_number(n));
            translate_simple_reconfiguration(nf, tree, vars, indices);
            i++;
        }
        indices = g_slist_remove(indices, cip);

        // n transitions
        // x equal to some constant, y different from all constants
        len = g_slist_length(indices) - 1; // last one is INDEX_OTHER
        cip = (struct case_index *) malloc(sizeof(struct case_index));
        assert(cip);
        cip->type = INDEX_VAR_X;
        strcpy(cip->orig_name, var_y);
        indices = g_slist_insert_sorted(indices, cip, compare_case_indexes);
        for (e = indices, i = 0; e; e = g_slist_next(e)) {
            if (((struct case_index *)e->data)->type != INDEX_CONST)
                continue;
            n1 = new_node(IN,
                    new_node(ID, (node_ptr) var_y, Nil),
                    new_node(ID, (node_ptr) set_y, Nil));
            if (set_y_compl)
                n1 = new_node(NOT, n1, Nil);

            n = new_number_node(((struct case_index *) e->data)->constant);
            n2 = new_node(IN, n, new_node(ID, (node_ptr) set_x, Nil));
            if (set_x_compl)
                n2 = new_node(NOT, n2, Nil);
            nf = rewrite_formula_replacing_var(node, var_x, n);
            nf = new_node(AND, n1, nf);
            nf = new_node(AND, n2, nf);

            for (ej = indices; ej; ej = g_slist_next(ej)) {
                if (((struct case_index *)ej->data)->type == INDEX_CONST) {
                    n1 = new_node(NOTEQUAL,
                            new_node(ID, (node_ptr) var_y, Nil),
                            new_number_node(((struct case_index *)
                                    ej->data)->constant));
                    nf = new_node(AND, n1, nf);
                }
            }

            printf("\n");
            printf(":comment transition t%d\n", ++mcmt_trans_count);
            printf(":comment reconfiguration %d, case 3,"
                    " transition %d of %d, %s -> %d, %s -> x\n", reconf_count,
                    len * len + len + i + 1, len * len + 2 * len + 2, var_x,
                    node_get_number(n), var_y);
            translate_simple_reconfiguration(nf, tree, vars, indices);
            i++;
        }
        indices = g_slist_remove(indices, cip);

        // second to last transition
        // x and y different from all constants and equal
        cip = (struct case_index *) malloc(sizeof(struct case_index));
        assert(cip);
        cip->type = INDEX_VAR_X;
        strcpy(cip->orig_name, var_x);
        indices = g_slist_insert_sorted(indices, cip, compare_case_indexes);
        n1 = new_node(IN,
                      new_node(ID, (node_ptr) var_x, Nil),
                      new_node(ID, (node_ptr) set_x, Nil));
        if (set_x_compl)
            n1 = new_node(NOT, n1, Nil);
        n2 = new_node(IN,
                      new_node(ID, (node_ptr) var_x, Nil),
                      new_node(ID, (node_ptr) set_y, Nil));
        if (set_y_compl)
            n2 = new_node(NOT, n2, Nil);
        nf = rewrite_formula_replacing_var(node, var_y,
                new_node(ID, (node_ptr) var_x, Nil));
        nf = new_node(AND, n2, nf);
        nf = new_node(AND, n1, nf);

        for (e = indices; e; e = g_slist_next(e)) {
            if (((struct case_index *)e->data)->type == INDEX_CONST) {
                n1 = new_node(NOTEQUAL,
                              new_node(ID, (node_ptr) var_x, Nil),
                              new_number_node(((struct case_index *)
                                               e->data)->constant));
                nf = new_node(AND, n1, nf);
            }
        }

        printf("\n");
        printf(":comment transition t%d\n", ++mcmt_trans_count);
        printf(":comment reconfiguration %d, case 3,"
                " transition %d of %d, %s -> x, %s -> x\n", reconf_count,
                len * len + 2 * len + 1, len * len + 2 * len + 2, var_x, var_y);
        translate_simple_reconfiguration(nf, tree, vars, indices);

        // last transition
        // x and y different from all constants and not equal
        cip = (struct case_index *) malloc(sizeof(struct case_index));
        assert(cip);
        cip->type = INDEX_VAR_Y;
        strcpy(cip->orig_name, var_y);
        indices = g_slist_insert_sorted(indices, cip, compare_case_indexes);
        n1 = new_node(IN,
                      new_node(ID, (node_ptr) var_x, Nil),
                      new_node(ID, (node_ptr) set_x, Nil));
        if (set_x_compl)
            n1 = new_node(NOT, n1, Nil);
        n2 = new_node(IN,
                      new_node(ID, (node_ptr) var_y, Nil),
                      new_node(ID, (node_ptr) set_y, Nil));
        if (set_y_compl)
            n2 = new_node(NOT, n2, Nil);
        nf = new_node(AND, n2, node);
        nf = new_node(AND, n1, nf);

        for (e = indices; e; e = g_slist_next(e)) {
            if (((struct case_index *)e->data)->type == INDEX_CONST) {
                n1 = new_node(NOTEQUAL,
                              new_node(ID, (node_ptr) var_x, Nil),
                              new_number_node(((struct case_index *)
                                               e->data)->constant));
                nf = new_node(AND, n1, nf);
            }
        }

        for (e = indices; e; e = g_slist_next(e)) {
            if (((struct case_index *)e->data)->type == INDEX_CONST) {
                n1 = new_node(NOTEQUAL,
                              new_node(ID, (node_ptr) var_y, Nil),
                              new_number_node(((struct case_index *)
                                               e->data)->constant));
                nf = new_node(AND, n1, nf);
            }
        }

        printf("\n");
        printf(":comment transition t%d\n", ++mcmt_trans_count);
        printf(":comment reconfiguration %d, case 3,"
                " transition %d of %d, %s -> x, %s -> y\n", reconf_count,
                len * len + 2 * len + 2, len * len + 2 * len + 2, var_x, var_y);
        translate_simple_reconfiguration(nf, tree, vars, indices);
    }

    g_array_free(invars, TRUE);
    g_array_free(comp_guards, TRUE);
    g_array_free(arr_indices, TRUE);
    e = indices;
    while (e) {
        free(e->data);
        e = g_slist_next(e);
    }
    g_slist_free(indices);
}

void find_component_guards(node_ptr node, GArray *guards)
{
    node_ptr n;

    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case SUBCOMPONENT:
            n = node_get_left(node_get_right(node));
            g_array_append_val(guards, n);
            break;
        case ID:
            // fall through
        case NUMBER:
            // fall through
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            break;
        default:
            find_component_guards(node_get_right(node), guards);
            find_component_guards(node_get_left(node), guards);
            break;
    }
}

gint compare_case_indexes(gconstpointer ci1, gconstpointer ci2)
{
    struct case_index *cidx1 = (struct case_index *) ci1;
    struct case_index *cidx2 = (struct case_index *) ci2;

    if (cidx1->type < cidx2->type)
        return -1;
    else if (cidx1->type > cidx2->type)
        return 1;
    // cidx1->type == cidx2->type

    if (cidx1->type == INDEX_VAR_X || cidx1->type == INDEX_VAR_Y)
        return strcmp(cidx1->orig_name, cidx2->orig_name);
    else if (cidx1->type == INDEX_CONST)
        return cidx1->constant - cidx2->constant;
    else
        return 0;
}

void find_reconf_different_indexes(node_ptr node, GArray *vars,
        GArray *indexes, int in_set)
{
    node_ptr idx, n;
    int i;

    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case INDEXED:
            find_reconf_different_indexes(node_get_left(node_get_right(node)),
                    vars, indexes, 1);
            break;
        case ADD:
            // fall through
        case REMOVE:
            find_reconf_different_indexes(node_get_right(node), vars,
                    indexes, 1);
            break;
        case IN:
            find_reconf_different_indexes(node_get_left(node), vars,
                    indexes, 1);
            break;
        case ID:
            // fall through
        case NUMBER:
            if (in_set) {
                idx = node;
                for (i = 0; i < indexes->len; i++) {
                    n = g_array_index(indexes, node_ptr, i);
                    if (indexes_equal(n, idx))
                        break;
                }
                if (i == indexes->len)
                    g_array_append_val(indexes, idx);
            }
            break;
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            break;
        default:
            find_reconf_different_indexes(node_get_left(node), vars,
                    indexes, in_set);
            find_reconf_different_indexes(node_get_right(node), vars,
                    indexes, in_set);
    }
}

node_ptr rewrite_formula_replacing_var(node_ptr node, char *x, node_ptr n)
{
    char *c;

    if (node == NULL)
        return NULL;

    switch(node_get_type(node)) {
        case ID:
            c = (char *) node_get_left(node);
            if (!strcmp(c, x))
                return n;
            else
                return node;
            break;
        case NUMBER:
            return new_number_node(node_get_number(node));
            break;
        default:
            return new_node(node_get_type(node),
                    rewrite_formula_replacing_var(node_get_left(node), x, n),
                    rewrite_formula_replacing_var(node_get_right(node), x, n));
            break;
    }
}

void translate_simple_reconfiguration(node_ptr node, node_ptr tree,
        GArray *vars, GSList *case_indices)
{
    GSList *cases = NULL; // elements are of type struct reconf_case
    GSList *e;
    struct reconf_case *tmp_case;
    GArray *conjs = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    struct case_index *cix = NULL, *ciy = NULL;
    int i;
    node_ptr n, repl_x = NULL, repl_y = NULL, par, val;
    GString *nullval = NULL;

    e = case_indices;
    while (e) {
        tmp_case = (struct reconf_case *) malloc(sizeof(struct reconf_case));
        tmp_case->index = (struct case_index *) e->data;
        tmp_case->updates = g_array_new(FALSE, FALSE, sizeof(GString *));
        for (i = 0; i < vars->len; i++)
            g_array_append_val(tmp_case->updates, nullval);

        cases = g_slist_prepend(cases, tmp_case);
        if (tmp_case->index->type == INDEX_VAR_X) {
            cix = tmp_case->index;
            repl_x = new_node(ID, (node_ptr) cix->orig_name, NULL);
        } if (tmp_case->index->type == INDEX_VAR_Y) {
            ciy = tmp_case->index;
            repl_y = new_node(ID, (node_ptr) ciy->orig_name, NULL);
        }
        e = g_slist_next(e);
    }
    cases = g_slist_reverse(cases);

    printf(":transition\n");
    printf(":var j\n");
    e = case_indices;
    if (cix)
        printf(":var x\n");
    if (ciy)
        printf(":var y\n");
    printf(":guard true");

    find_conjuncts(node, conjs);
    for (i = 0; i < conjs->len; i++) {
        n = g_array_index(conjs, node_ptr, i);
        if (!is_update_atom(n)) {
            printf(" ");
            translate_atom_replacing_ids_and_nums(n, vars, repl_x, "x",
                    repl_y, "y", NULL);
        } else {
            if (node_get_type(node_get_left(n)) == NEXT) {
                par = node_get_left(node_get_left(n));
                val = node_get_right(n);
            } else {
                par = node_get_left(node_get_right(n));
                val = node_get_left(n);
            }

            switch (node_get_type(val)) {
                case ADD:
                    translate_reconf_add_remove(1, par, node_get_right(val),
                            vars, cases, repl_x, "x", repl_y, "y");
                    break;
                case REMOVE:
                    translate_reconf_add_remove(0, par, node_get_right(val),
                            vars, cases, repl_x, "x", repl_y, "y");
                    break;
                default:
                    translate_reconf_par_update(par, val, vars, cases,
                            repl_x, "x", repl_y, "y");
                    break;
            }
        }
    }

    translate_reconf_unset_parameters(vars, cases, repl_x, "x", repl_y, "y");

    translate_reconf_unset_tokens(tree, vars, cases);

    translate_reconf_invars(tree, vars, cases);

    translate_reconf_cases(cases);

    e = cases;
    while (e) {
        g_array_free(((struct reconf_case *) e->data)->updates, FALSE);
        e = g_slist_next(e);
    }
    g_slist_free(cases);

    g_array_free(conjs, TRUE);
}

int is_update_atom(node_ptr node)
{
    node_ptr l, r;

    if (node == NULL || node_get_type(node) != EQUAL)
        return 0;

    l = node_get_left(node);
    r = node_get_right(node);

    return ((l && node_get_type(l) == NEXT) || (r && node_get_type(r) == NEXT));
}

void translate_reconf_add_remove(int is_add, node_ptr par, node_ptr val,
        GArray* vars, GSList *cases,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y)
{
    GArray *elems = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    int i;
    node_ptr n;
    GString *u_value;
    int j;
    GSList *e;
    struct reconf_case *rc;
    struct mcmt_var v, is;
    struct case_index *ci;

    get_elements_of_set(val, elems);
    for (i = 0; i < elems->len; i++) {
        n = (node_ptr) g_array_index(elems, node_ptr, i);
        printf(" (= ");
        translate_id(par, vars, NULL, NULL, NULL);
        printf("[");
        if (node_get_type(n) == NUMBER)
            printf("%d", node_get_number(n));
        else
            translate_term_replacing_ids_and_nums(n, vars, repl_x, x,
                    repl_y, y, NULL);
        printf("]");
        if (is_add)
            printf(" false)");
        else
            printf(" true)");

        j = get_index_of_var(vars, (char *) node_get_left(par));
        v = g_array_index(vars, struct mcmt_var, j);
        ci = get_case_index_for_node(n, repl_x, repl_y, cases);
        u_value = g_string_new(NULL);
        if (is_add)
            g_string_printf(u_value, "true");
        else
            g_string_printf(u_value, "false");
        set_update_val(v.mcmt_name, ci, u_value, cases, vars);

        for (j = 0; j < vars->len; j++) {
            v = (struct mcmt_var) g_array_index(vars, struct mcmt_var, j);
            if (v.kind == KIND_TOKEN || v.kind == KIND_TOKEN_SET) {
                is = (struct mcmt_var) g_array_index(vars, struct mcmt_var,
                        v.index_set);
                if (!strcmp(is.orig_name, (char *) node_get_left(par)) &&
                        !v.is_src) {
                    ci = get_case_index_for_node(n, repl_x, repl_y, cases);
                    u_value = g_string_new(NULL);
                    g_string_printf(u_value, "false");
                    set_update_val(v.mcmt_name, ci, u_value, cases, vars);
                }
            }
        }
    }

    i = get_index_of_var(vars, (char *) node_get_left(par));
    is = (struct mcmt_var) g_array_index(vars, struct mcmt_var, i);
    if (is.size_var != -1) {
        v = (struct mcmt_var) g_array_index(vars, struct mcmt_var, is.size_var);
        e = cases;
        while (e) {
            rc = (struct reconf_case *) e->data;
            u_value = g_string_new(NULL);
            if (is_add)
                g_string_printf(u_value, "(+ %s %d)", v.mcmt_name, elems->len);
            else
                g_string_printf(u_value, "(+ %s (* -1 %d))", v.mcmt_name,
                        elems->len);
            set_update_val(v.mcmt_name, rc->index, u_value, cases, vars);
            e = g_slist_next(e);
        }
    }

    g_array_free(elems, TRUE);
}

void translate_reconf_par_update(node_ptr par, node_ptr val,
        GArray* vars, GSList *cases,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y)
{
    node_ptr id, idx;
    int i;
    struct mcmt_var v;
    struct case_index *ci;
    GString *u_value;
    GSList *e;
    struct reconf_case *rc;

    if (node_get_type(par) == ID) {
        i = get_index_of_var(vars, (char *) node_get_left(par));
        v = (struct mcmt_var) g_array_index(vars, struct mcmt_var, i);

        u_value = g_string_new(NULL);
        translate_term_replacing_ids_and_nums(val, vars, repl_x, x,
                repl_y, y, u_value);

        e = cases;
        while (e) {
            rc = (struct reconf_case *) e->data;
            set_update_val(v.mcmt_name, rc->index, u_value, cases, vars);
            e = g_slist_next(e);
        }
    } else if (node_get_type(par) == INDEXED) {
        id = node_get_left(par);
        i = get_index_of_var(vars, (char *) node_get_left(id));
        v = (struct mcmt_var) g_array_index(vars, struct mcmt_var, i);

        idx = node_get_left(node_get_right(par));
        ci = get_case_index_for_node(idx, repl_x, repl_y, cases);

        u_value = g_string_new(NULL);
        translate_term_replacing_ids_and_nums(val, vars, repl_x, x,
                repl_y, y, u_value);
        set_update_val(v.mcmt_name, ci, u_value, cases, vars);
    } else {
        printf("\n");
        fprintf(stderr, "invalid parameter in reconfiguration\n");
        exit(1);
    }
}

void translate_reconf_unset_parameters(GArray* vars, GSList *cases,
        node_ptr repl_x, char *x, node_ptr repl_y, char *y)
{
    int i;
    struct mcmt_var v;
    GSList *e;
    struct reconf_case *rc;
    GString *u_value;
    GString **s;

    for (i = 0; i < vars->len; i++) {
        v = (struct mcmt_var) g_array_index(vars, struct mcmt_var, i);
        if (v.kind != KIND_TOKEN && v.kind != KIND_TOKEN_SET) {
            e = cases;
            while (e) {
                rc = (struct reconf_case *) e->data;
                s = &g_array_index(rc->updates, GString *, i);
                if (*s == NULL) {
                    u_value = g_string_new(NULL);
                    if (v.kind == KIND_INDEX_SET || v.kind == KIND_PARAM_SET) {
                        if (rc->index->type != INDEX_CONST)
                            g_string_printf(u_value, "%s[j]", v.mcmt_name);
                        else
                            g_string_printf(u_value, "%s[%d]", v.mcmt_name,
                                    rc->index->constant);
                    } else {
                        g_string_printf(u_value, "%s", v.mcmt_name);
                    }

                    set_update_val(v.mcmt_name, rc->index, u_value, cases,
                            vars);
                }
                e = g_slist_next(e);
            }
        }
    }
}

void translate_reconf_unset_tokens(node_ptr tree, GArray* vars, GSList *cases)
{
    int i;
    struct mcmt_var v;
    GSList *e;
    struct reconf_case *rc;
    GString *u_value;
    GString **s;
    node_ptr guard;

    for (i = 0; i < vars->len; i++) {
        v = (struct mcmt_var) g_array_index(vars, struct mcmt_var, i);
        if (v.kind == KIND_TOKEN || v.kind == KIND_TOKEN_SET) {
            e = cases;
            while (e) {
                rc = (struct reconf_case *) e->data;
                s = &g_array_index(rc->updates, GString *, i);
                if (*s == NULL) {
                    u_value = g_string_new(NULL);
                    if (!v.is_src) {
                        if (v.kind == KIND_TOKEN)
                            g_string_append_printf(u_value, "(and %s ",
                                    v.mcmt_name);
                        else
                            g_string_append_printf(u_value, "(and %s[j] ",
                                    v.mcmt_name);
                        guard = find_subcomp_guard(tree, v.orig_name);
                        translate_conjunction_using_next_values(1, guard, vars,
                                cases, u_value);
                        g_string_append_printf(u_value, ")");
                    } else {
                        g_string_printf(u_value, "true");
                    }

                    set_update_val(v.mcmt_name, rc->index, u_value, cases,
                            vars);
                }

                e = g_slist_next(e);
            }
        }
    }
}

void translate_reconf_invars(node_ptr tree, GArray* vars, GSList *cases)
{
    int i;
    node_ptr n;
    GArray *invars = g_array_new(FALSE, FALSE, sizeof(node_ptr));
    find_invariant_formulas(tree, invars, 0);
    for (i = 0; i < invars->len; i++) {
        n = g_array_index(invars, node_ptr, i);
        printf(" ");
        translate_conjunction_using_next_values(0, n, vars, cases, NULL);
    }
    printf("\n");

    g_array_free(invars, TRUE);
}

void translate_conjunction_using_next_values(int structure, node_ptr node,
        GArray* vars, GSList *cases, GString *out)
{
    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case AND:
            if (structure)
                printf_redir(out, " (and");
            translate_conjunction_using_next_values(structure,
                    node_get_left(node), vars, cases, out);
            translate_conjunction_using_next_values(structure,
                    node_get_right(node), vars, cases, out);
            if (structure)
                printf_redir(out, ")");
            break;
        case NOT:
            translate_in_using_next_values(node_get_left(node), vars, cases, 0,
                    out);
            break;
        case IN:
            translate_in_using_next_values(node, vars, cases, 1, out);
            break;
        case LE:
            // fall through
        case GE:
            // fall through
        case LT:
            // fall through
        case GT:
            // fall through
        case EQUAL:
            // fall through
        case NOTEQUAL:
            translate_atom_using_next_values(node, vars, cases, out);
            break;
        case ID:
            // fall through
        case NUMBER:
            break;
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            translate_atom_using_next_values(node, vars, cases, out);
            break;
        default:
            break;
    }
}

struct reconf_case *get_reconf_case_for_const_index_node(node_ptr idx,
        GSList *cases)
{
    GSList *e;
    struct reconf_case *rc;
    int num;

    assert(node_get_type(idx) == NUMBER);
    num = node_get_number(idx);

    e = cases;
    while (e) {
        rc = (struct reconf_case *) e->data;
        if (rc->index->type == INDEX_CONST && rc->index->constant == num)
            break;
        if (rc->index->type == INDEX_OTHER)
            break;
        e = g_slist_next(e);
    }
    assert(e);

    return rc;
}

void translate_in_using_next_values(node_ptr node, GArray* vars, GSList *cases,
        int val, GString *out)
{
    node_ptr term;
    node_ptr set;
    int i;
    struct reconf_case *rc;
    GString *next;

    term = node_get_left(node);
    set = node_get_right(node);

    rc = get_reconf_case_for_const_index_node(term, cases);
    assert(rc->index->type == INDEX_CONST);

    i = get_index_of_var(vars, (char *) node_get_left(set));
    next = g_array_index(rc->updates, GString *, i);

    printf_redir(out, " (= %s", next->str);
    if (val)
        printf_redir(out, " true)");
    else
        printf_redir(out, " false)");
}

void translate_atom_using_next_values(node_ptr node, GArray* vars,
        GSList *cases, GString *out)
{
    node_ptr id, idx;
    int num;

    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case LE:
            printf_redir(out, " (<= ");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, ")");
            break;
        case GE:
            printf_redir(out, " (>= ");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, ")");
            break;
        case LT:
            printf_redir(out, " (< ");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, ")");
            break;
        case GT:
            printf_redir(out, " (> ");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, ")");
            break;
        case EQUAL:
            printf_redir(out, " (= ");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, ")");
            break;
        case NOTEQUAL:
            printf_redir(out, " (not (=");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, "))");
            break;
        case PLUS:
            printf_redir(out, " (+ ");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, ")");
            break;
        case MINUS:
            printf_redir(out, " (+ ");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " (* -1 ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, "))");
            break;
        case TIMES:
            printf_redir(out, " (* ");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, ")");
            break;
        case FRAC:
            printf_redir(out, " (div ");
            translate_atom_using_next_values(node_get_left(node), vars, cases,
                    out);
            printf_redir(out, " ");
            translate_atom_using_next_values(node_get_right(node), vars, cases,
                    out);
            printf_redir(out, ")");
            break;
        case SIZE:
            translate_size_using_next_values(node, vars, cases, out);
            break;
        case ID:
            translate_id_using_next_values(node, vars, cases, out);
            break;
        case NUMBER:
            num = node_get_number(node);
            printf_redir(out, "%d", num);
            break;
        case BOOL_TRUE:
            printf_redir(out, "true");
            break;
        case BOOL_FALSE:
            printf_redir(out, "false");
            break;
        case INDEXED:
            id = node_get_left(node);
            idx = node_get_left(node_get_right(node));
            translate_id(id, vars, NULL, NULL, out);
            printf_redir(out, "[");
            translate_atom_using_next_values(idx, vars, cases, out);
            printf_redir(out, "]");
            break;
        default:
            break;
    }
}

void translate_size_using_next_values(node_ptr node, GArray* vars,
        GSList *cases, GString *out)
{
    int i;
    struct mcmt_var v;
    struct reconf_case *rc;
    GString *s;

    i = get_index_of_var(vars, (char *) node_get_left(node_get_left(node)));
    if (i == -1) {
        printf("\n");
        fprintf(stderr, "cannot translate identifier %s\n",
                (char *) node_get_left(node_get_left(node)));
        exit(1);
    }
    v = g_array_index(vars, struct mcmt_var, i);
    rc = (struct reconf_case *) cases->data;
    s = g_array_index(rc->updates, GString *, v.size_var);
    printf_redir(out, "%s", s->str);
}

void translate_id_using_next_values(node_ptr node, GArray* vars,
        GSList *cases, GString *out)
{
    int i;
    struct reconf_case *rc;
    GString *s;

    i = get_index_of_var(vars, (char *) node_get_left(node));
    if (i == -1) {
        printf("\n");
        fprintf(stderr, "cannot translate identifier %s\n",
                (char *) node_get_left(node_get_left(node)));
        exit(1);
    }
    rc = (struct reconf_case *) cases->data;
    s = g_array_index(rc->updates, GString *, i);
    printf_redir(out, "%s", s->str);
}

void translate_reconf_cases(GSList *cases)
{
    int num_cases = g_slist_length(cases);
    GSList *e, *e2;
    struct reconf_case *rc, *rc2;
    int has_x = 0, has_y = 0;
    int i;
    GString *s;

    printf(":numcases %d\n", num_cases);

    e = cases;
    while (e) {
        rc = (struct reconf_case *) e->data;

        printf(":case");
        if (num_cases == 1) {
            printf(" true\n");
        } else if (rc->index->type == INDEX_VAR_X) {
            printf(" (= x j)\n");
            has_x = 1;
        } else if (rc->index->type == INDEX_VAR_Y) {
            printf(" (not (= x j)) (= y j)\n");
            has_y = 1;
        } else {
            if (has_x)
                printf(" (not (= x j))");
            if (has_y)
                printf(" (not (= y j))");
            if (rc->index->type == INDEX_CONST) {
                printf(" (= %d j)\n", rc->index->constant);
            } else {
                e2 = cases;
                while (e2) {
                    rc2 = (struct reconf_case *) e2->data;
                    if (rc2->index->type == INDEX_CONST)
                        printf(" (not (= %d j))", rc2->index->constant);
                    e2 = g_slist_next(e2);
                }
                printf("\n");
            }
        }

        for (i = 0; i < rc->updates->len; i++) {
            s = g_array_index(rc->updates, GString *, i);
            assert(s != NULL);
            printf(":val %s\n", s->str);
        }

        e = g_slist_next(e);
    }
}

void get_elements_of_set(node_ptr node, GArray *elems)
{
    int i;
    node_ptr n;

    if (node == NULL)
        return;

    switch(node_get_type(node)) {
        case ID:
            // fall through
        case NUMBER:
            // fall through
            for (i = 0; i < elems->len; i++) {
                n = (node_ptr) g_array_index(elems, node_ptr, i);
                if (indexes_equal(node, n))
                    break;
            }
            if (i == elems->len)
                g_array_append_val(elems, node);
            break;
        case BOOL_TRUE:
            // fall through
        case BOOL_FALSE:
            assert(0);
            break;
        default:
            get_elements_of_set(node_get_right(node), elems);
            get_elements_of_set(node_get_left(node), elems);
            break;
    }
}

void set_update_val(char *name, struct case_index *ci, GString *value,
        GSList *cases, GArray *vars)
{
    GSList *e;
    struct reconf_case *rc;
    struct case_index *ci2;
    int i;
    GString **s;

    /* DEBUG
    GString *c = g_string_new(NULL);
    if (ci->type == INDEX_OTHER)
        g_string_printf(c, "other");
    else if (ci->type == INDEX_VAR_X)
        g_string_printf(c, "x");
    else if (ci->type == INDEX_VAR_Y)
        g_string_printf(c, "y");
    else
        g_string_printf(c, "%d", ci->constant);
    printf("update(%s in case %s to value %s)\n", name, c->str, value->str);
    // END DEBUG */

    e = cases;
    while (e) {
        rc = (struct reconf_case *) e->data;
        ci2 = rc->index;

        if ((ci->type == ci2->type)
            && (ci->type == INDEX_OTHER
                || ci->type == INDEX_VAR_X
                || ci->type == INDEX_VAR_Y
                || (ci->type == INDEX_CONST && ci->constant == ci2->constant)))
            break;

        e = g_slist_next(e);
    }
    assert(e);

    i = get_index_of_var_mcmt_name(vars, name);
    assert(i != -1);
    s = &g_array_index(rc->updates, GString *, i);
    assert(*s == NULL);
    *s = value;
}

struct case_index *get_case_index_for_node(node_ptr n,
        node_ptr repl_x, node_ptr repl_y, GSList *cases)
{
    GSList *e;
    struct reconf_case *rc;
    struct case_index *ci;

    e = cases;
    while (e) {
        rc = (struct reconf_case *) e->data;
        ci = rc->index;

        if ((ci->type == INDEX_VAR_X && indexes_equal(n, repl_x)) ||
            (ci->type == INDEX_VAR_Y && indexes_equal(n, repl_y)) ||
            (ci->type == INDEX_CONST && node_get_type(n) == NUMBER &&
             ci->constant == node_get_number(n)))
            break;

        e = g_slist_next(e);
    }

    if (e)
        return ci;
    else
        return NULL;
}
